"""Utilities to work with Jedi

Translates pygls types back and forth with Jedi
"""

from typing import Dict, Optional

import jedi.api.errors
from jedi import Project, Script
from jedi.api.classes import Name
from jedi.api.environment import get_cached_default_environment
from pygls.types import (
    Diagnostic,
    DiagnosticSeverity,
    Location,
    Position,
    Range,
    SymbolInformation,
)
from pygls.uris import from_fs_path
from pygls.workspace import Workspace

from .type_map import get_lsp_symbol_type


def script(workspace: Workspace, uri: str) -> Script:
    """Simplifies getting jedi Script"""
    project_ = project(workspace)
    document = workspace.get_document(uri)
    return Script(
        code=document.source,
        path=document.path,
        project=project_,
        environment=get_cached_default_environment(),
    )


def project(workspace: Workspace) -> Project:
    """Simplifies getting jedi project"""
    return Project(
        path=workspace.root_path,
        smart_sys_path=True,
        load_unsafe_extensions=False,
    )


def lsp_range(name: Name) -> Range:
    """Get LSP range from Jedi definition

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, subtract 1 from Jedi's definition line
    """
    return Range(
        start=Position(line=name.line - 1, character=name.column),
        end=Position(
            line=name.line - 1, character=name.column + len(name.name),
        ),
    )


def lsp_location(name: Name) -> Location:
    """Get LSP location from Jedi definition"""
    return Location(uri=from_fs_path(name.module_path), range=lsp_range(name))


def lsp_symbol_information(name: Name) -> SymbolInformation:
    """Get LSP SymbolInformation from Jedi definition"""
    return SymbolInformation(
        name=name.name,
        kind=get_lsp_symbol_type(name.type),
        location=lsp_location(name),
        container_name=parent_name(name),
    )


def lsp_diagnostic(error: jedi.api.errors.SyntaxError) -> Diagnostic:
    """Get LSP Diagnostic from Jedi SyntaxError"""
    return Diagnostic(
        range=Range(
            start=Position(line=error.line - 1, character=error.column),
            end=Position(line=error.line - 1, character=error.column),
        ),
        message=str(error).strip("<>"),
        severity=DiagnosticSeverity.Error,
        source="jedi",
    )


def parent_name(name: Name) -> Optional[str]:
    """Retrieve the parent name from Jedi

    Error prone Jedi calls are wrapped in try/except to avoid
    """
    try:
        parent = name.parent()
    except Exception:  # pylint: disable=broad-except
        return None
    return parent.name if parent and parent.parent() else None


def line_column(position: Position) -> Dict[str, int]:
    """Translate pygls Position to Jedi's line / column

    Returns a dictionary because this return result should be unpacked as a
    function argument to Jedi's functions.

    Jedi is 1-indexed for lines and 0-indexed for columns. LSP is 0-indexed for
    lines and 0-indexed for columns. Therefore, add 1 to LSP's request for the
    line.
    """
    return dict(line=position.line + 1, column=position.character)


def compare_names(name: Name, name_root: Name) -> bool:
    """Check if a name has the same root as another name

    Assumes the first result is correct. This could potentially be the cause of
    bugs; if you experience weird behavior we may want to revisit this
    decision.

    Currently useful for "highlight" functionality
    """
    names = name.goto(follow_imports=False, follow_builtin_imports=False)
    return names and names[0] == name_root



"""Jedi types mapped to LSP types"""

from pygls.types import CompletionItemKind, SymbolKind

# Taken from: https://github.com/palantir/python-language-server
_JEDI_COMPLETION_TYPE_MAP = {
    "buffer": CompletionItemKind.Class,
    "builtin": CompletionItemKind.Class,
    "builtinfunction": CompletionItemKind.Function,
    "class": CompletionItemKind.Class,
    "constant": CompletionItemKind.Variable,
    "dict": CompletionItemKind.Class,
    "dictionary": CompletionItemKind.Class,
    "dictproxy": CompletionItemKind.Class,
    "file": CompletionItemKind.File,
    "frame": CompletionItemKind.Class,
    "funcdef": CompletionItemKind.Function,
    "function": CompletionItemKind.Function,
    "generator": CompletionItemKind.Function,
    "import": CompletionItemKind.Module,
    "instance": CompletionItemKind.Reference,
    "keyword": CompletionItemKind.Keyword,
    "lambda": CompletionItemKind.Function,
    "method": CompletionItemKind.Method,
    "module": CompletionItemKind.Module,
    "none": CompletionItemKind.Value,
    "param": CompletionItemKind.Variable,
    "property": CompletionItemKind.Property,
    "slice": CompletionItemKind.Class,
    "statement": CompletionItemKind.Keyword,
    "traceback": CompletionItemKind.Class,
    "tuple": CompletionItemKind.Class,
    "type": CompletionItemKind.Class,
    "value": CompletionItemKind.Value,
    "variable": CompletionItemKind.Variable,
    "xrange": CompletionItemKind.Class,
}

_JEDI_SYMBOL_TYPE_MAP = {
    "none": SymbolKind.Variable,
    "type": SymbolKind.Class,
    "tuple": SymbolKind.Class,
    "dict": SymbolKind.Class,
    "dictionary": SymbolKind.Class,
    "function": SymbolKind.Function,
    "lambda": SymbolKind.Function,
    "generator": SymbolKind.Function,
    "class": SymbolKind.Class,
    "instance": SymbolKind.Class,
    "method": SymbolKind.Method,
    "builtin": SymbolKind.Class,
    "builtinfunction": SymbolKind.Function,
    "module": SymbolKind.Module,
    "file": SymbolKind.File,
    "xrange": SymbolKind.Array,
    "slice": SymbolKind.Class,
    "traceback": SymbolKind.Class,
    "frame": SymbolKind.Class,
    "buffer": SymbolKind.Array,
    "dictproxy": SymbolKind.Class,
    "funcdef": SymbolKind.Function,
    "property": SymbolKind.Property,
    "import": SymbolKind.Module,
    "keyword": SymbolKind.Variable,
    "constant": SymbolKind.Constant,
    "variable": SymbolKind.Variable,
    "value": SymbolKind.Variable,
    "param": SymbolKind.Variable,
    "statement": SymbolKind.Variable,
    "boolean": SymbolKind.Boolean,
    "int": SymbolKind.Number,
    "longlean": SymbolKind.Number,
    "float": SymbolKind.Number,
    "complex": SymbolKind.Number,
    "string": SymbolKind.String,
    "unicode": SymbolKind.String,
    "list": SymbolKind.Array,
}


def get_lsp_completion_type(jedi_type: str) -> CompletionItemKind:
    """Get type map. Always return a value."""
    return _JEDI_COMPLETION_TYPE_MAP.get(jedi_type, CompletionItemKind.Text)


def get_lsp_symbol_type(jedi_type: str) -> SymbolKind:
    """Get type map. Always return a value."""
    return _JEDI_SYMBOL_TYPE_MAP.get(jedi_type, SymbolKind.Namespace)



"""Jedi Language Server

Creates the language server constant and wraps "features" with it.

Official language server spec:
    https://microsoft.github.io/language-server-protocol/specification
"""

import itertools
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional

from pygls.exceptions import FeatureAlreadyRegisteredError
from pygls.features import (
    COMPLETION,
    DEFINITION,
    DOCUMENT_HIGHLIGHT,
    DOCUMENT_SYMBOL,
    HOVER,
    INITIALIZED,
    REFERENCES,
    RENAME,
    SIGNATURE_HELP,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    WORKSPACE_SYMBOL,
)
from pygls.server import LanguageServer
from pygls.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    ConfigurationItem,
    ConfigurationParams,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    DocumentHighlight,
    DocumentSymbolParams,
    Hover,
    Location,
    ParameterInformation,
    Registration,
    RegistrationParams,
    RenameParams,
    SignatureHelp,
    SignatureInformation,
    SymbolInformation,
    TextDocumentPositionParams,
    TextEdit,
    Unregistration,
    UnregistrationParams,
    WorkspaceEdit,
    WorkspaceSymbolParams,
)

from . import jedi_utils, pygls_utils
from .pygls_utils import Feature, FeatureConfig, rgetattr
from .type_map import get_lsp_completion_type


class JediLanguageServer(LanguageServer):
    """The Jedi Language Server

    :attr lookup_feature_id: map feature name to feature id.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lookup_feature_id: DefaultDict[str, str] = defaultdict(
            pygls_utils.uuid
        )

    def jls_message(self, message: str) -> None:
        """Standardize `JediLanguageServer.show_message`"""
        self.show_message(f"jedi-language-server {message}")

    async def unregister_feature(self, feature: Feature) -> None:
        """Unregister an LSP method if it has already been registered"""
        lsp_method = feature.lsp_method
        if lsp_method not in self.lookup_feature_id:
            return
        await self.unregister_capability_async(
            UnregistrationParams(
                [
                    Unregistration(
                        self.lookup_feature_id[lsp_method], lsp_method
                    )
                ]
            )
        )
        del self.lookup_feature_id[lsp_method]

    async def register_feature(
        self, feature: Feature, config: object,
    ) -> None:
        """(unregister and) register feature with new options

        :param config: full configuration object, used to lookup relevant
            feature configurations sent by the client
        """
        await self.unregister_feature(feature)
        registration_options = {
            cfg.path.split(".")[-1]: rgetattr(config, cfg.path, cfg.default)
            for cfg in feature.config
        }
        register_params = RegistrationParams(
            [
                Registration(
                    self.lookup_feature_id[feature.lsp_method],
                    feature.lsp_method,
                    registration_options,
                )
            ]
        )
        response = await self.register_capability_async(register_params)
        if response is None:
            try:
                self.feature(feature.lsp_method)(feature.function)
            except FeatureAlreadyRegisteredError:
                pass


def completion(
    server: JediLanguageServer, params: CompletionParams
) -> CompletionList:
    """Returns completion items"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    completions = jedi_script.complete(**jedi_lines)
    char = pygls_utils.char_before_cursor(
        document=server.workspace.get_document(params.textDocument.uri),
        position=params.position,
    )
    return CompletionList(
        is_incomplete=False,
        items=[
            CompletionItem(
                label=completion.name,
                kind=get_lsp_completion_type(completion.type),
                detail=completion.description,
                documentation=completion.docstring(),
                insert_text=pygls_utils.clean_completion_name(
                    completion.name, char
                ),
            )
            for completion in completions
        ],
    )


def signature_help(
    server: JediLanguageServer, params: TextDocumentPositionParams
) -> SignatureHelp:
    """Returns signature help"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    signatures = jedi_script.get_signatures(**jedi_lines)
    return SignatureHelp(
        signatures=[
            SignatureInformation(
                label=signature.to_string(),
                documentation=None,
                parameters=[
                    ParameterInformation(
                        label=info.to_string(), documentation=None
                    )
                    for info in signature.params
                ],
            )
            for signature in signatures
        ],
        active_signature=0,
        active_parameter=signatures[0].index if signatures else 0,
    )


def definition(
    server: JediLanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Support Goto Definition"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    names = jedi_script.goto(
        follow_imports=True, follow_builtin_imports=True, **jedi_lines,
    )
    return [jedi_utils.lsp_location(name) for name in names]


def highlight(
    server: JediLanguageServer, params: TextDocumentPositionParams
) -> List[DocumentHighlight]:
    """Support document highlight request"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    jedi_names = jedi_script.goto(
        follow_imports=False, follow_builtin_imports=False, **jedi_lines,
    )
    if not jedi_names:
        return []
    current_name = jedi_names[0]
    return [
        DocumentHighlight(jedi_utils.lsp_range(name))
        for name in jedi_script.get_names(
            all_scopes=True, definitions=True, references=True
        )
        if jedi_utils.compare_names(name, current_name)
    ]


def hover(
    server: JediLanguageServer, params: TextDocumentPositionParams
) -> Hover:
    """Support Hover"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    jedi_docstrings = (n.docstring() for n in jedi_script.help(**jedi_lines))
    names = [name for name in jedi_docstrings if name]
    return Hover(contents=names if names else "jedi: no help docs found")


def references(
    server: JediLanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Obtain all references to text"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    names = jedi_script.get_references(**jedi_lines)
    return [jedi_utils.lsp_location(name) for name in names]


def rename(
    server: JediLanguageServer, params: RenameParams
) -> Optional[WorkspaceEdit]:
    """Rename a symbol across a workspace"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    jedi_lines = jedi_utils.line_column(params.position)
    names = jedi_script.get_references(**jedi_lines)
    if not names:
        return None
    locations = [jedi_utils.lsp_location(name) for name in names]
    changes: Dict[str, List[TextEdit]] = {}
    for location in locations:
        text_edit = TextEdit(location.range, new_text=params.newName)
        if location.uri not in changes:
            changes[location.uri] = [text_edit]
        else:
            changes[location.uri].append(text_edit)
    return WorkspaceEdit(changes=changes)


def document_symbol(
    server: JediLanguageServer, params: DocumentSymbolParams
) -> List[SymbolInformation]:
    """Document Python document symbols"""
    jedi_script = jedi_utils.script(server.workspace, params.textDocument.uri)
    names = jedi_script.get_names()
    return [jedi_utils.lsp_symbol_information(name) for name in names]


def workspace_symbol(
    server: JediLanguageServer, params: WorkspaceSymbolParams
) -> List[SymbolInformation]:
    """Document Python workspace symbols"""
    jedi_project = jedi_utils.project(server.workspace)
    if params.query.strip() == "":
        return []
    names = jedi_project.search(params.query)
    return [
        jedi_utils.lsp_symbol_information(name)
        for name in itertools.islice(names, 20)
    ]


def _publish_diagnostics(server: JediLanguageServer, uri: str):
    """Helper function to publish diagnostics for a file"""
    jedi_script = jedi_utils.script(server.workspace, uri)
    errors = jedi_script.get_syntax_errors()
    diagnostics = [jedi_utils.lsp_diagnostic(error) for error in errors]
    server.publish_diagnostics(uri, diagnostics)


def did_save(server: JediLanguageServer, params: DidSaveTextDocumentParams):
    """Actions run on textDocument/didSave"""
    _publish_diagnostics(server, params.textDocument.uri)


def did_change(
    server: JediLanguageServer, params: DidChangeTextDocumentParams
):
    """Actions run on textDocument/didChange"""
    _publish_diagnostics(server, params.textDocument.uri)


def did_open(server: JediLanguageServer, params: DidOpenTextDocumentParams):
    """Actions run on textDocument/didOpen"""
    _publish_diagnostics(server, params.textDocument.uri)


# Feature constants


_CONFIG_COMPLETION = (
    FeatureConfig("completion.triggerCharacters", [".", "'", '"']),
)

_CONFIG_SIGNATURE = (
    FeatureConfig("signatureHelp.triggerCharacters", ["(", ",", ")"]),
)

_FEATURES_STANDARD = (
    Feature(COMPLETION, completion, _CONFIG_COMPLETION),
    Feature(DEFINITION, definition),
    Feature(DOCUMENT_HIGHLIGHT, highlight),
    Feature(DOCUMENT_SYMBOL, document_symbol),
    Feature(HOVER, hover),
    Feature(REFERENCES, references),
    Feature(RENAME, rename),
    Feature(SIGNATURE_HELP, signature_help, _CONFIG_SIGNATURE),
    Feature(WORKSPACE_SYMBOL, workspace_symbol),
)
_FEATURE_DID_CHANGE = Feature(TEXT_DOCUMENT_DID_CHANGE, did_change)
_FEATURE_DID_OPEN = Feature(TEXT_DOCUMENT_DID_OPEN, did_open)
_FEATURE_DID_SAVE = Feature(TEXT_DOCUMENT_DID_SAVE, did_save)


# Instantiate server singleton


SERVER = JediLanguageServer()


# Define post-initialization function to customize configuration


@SERVER.feature(INITIALIZED)
async def initialized(
    server: JediLanguageServer,
    params: Dict[str, object],  # pylint: disable=unused-argument
):
    """Post-basic initialization actions

    1. If server is not enabled, do not register any features
    2. Otherwise, register all features
    """
    _config = await server.get_configuration_async(
        ConfigurationParams([ConfigurationItem(section="jedi")])
    )
    config = _config[0] if _config else object()
    if not getattr(config, "enabled", True):
        server.jls_message("disabled; `jedi.enabled` set to `false`}")
        return
    for feature in _FEATURES_STANDARD:
        await server.register_feature(feature, config)

    if rgetattr(config, "diagnostics.enabled", True):
        if rgetattr(config, "diagnostics.didOpen", True):
            await server.register_feature(_FEATURE_DID_OPEN, config)
        else:
            await server.unregister_feature(_FEATURE_DID_OPEN)

        if rgetattr(config, "diagnostics.didChange", True):
            await server.register_feature(_FEATURE_DID_CHANGE, config)
        else:
            await server.unregister_feature(_FEATURE_DID_CHANGE)

        if rgetattr(config, "diagnostics.didSave", True):
            await server.register_feature(_FEATURE_DID_SAVE, config)
        else:
            await server.unregister_feature(_FEATURE_DID_SAVE)
    else:
        await server.unregister_feature(_FEATURE_DID_OPEN)
        await server.unregister_feature(_FEATURE_DID_CHANGE)
        await server.unregister_feature(_FEATURE_DID_SAVE)

    server.jls_message("initialized")



"""Utilities to work with pygls

Helper functions that simplify working with pygls
"""


import functools
from typing import Callable, NamedTuple, Tuple
from uuid import uuid4

from pygls.types import Position
from pygls.workspace import Document

_SENTINEL = object()


class FeatureConfig(NamedTuple):
    """Configuration item for a feature with associated defaults

    :attr arg: the name of the option within the feature
    :attr path: the lookup path, rooted at `jedi`, to the user configuration.
        The last item in the path must be the option associated with a feature
    :attr default: the value used if no value is found in the path

    Example:

        settings.json:

        {
          "jedi.completion.triggerCharacters": [".", "'", "\""]
        }

        python code associated with this configuration:

        FeatureConfig("completion.triggerCharacters", [".", "'", '"'])

    NOTE: all paths are rooted at "jedi". Do not specify the top-level jedi in
    the path.
    """

    path: str
    default: object


class Feature(NamedTuple):
    """Organize information about LanguageServer features

    :attr lsp_method: exact name of the feature's LSP method
    :attr function: python function to be associated with the method
    :attr config: collection of FeatureConfig supported by this feature
    """

    lsp_method: str
    function: Callable
    config: Tuple[FeatureConfig, ...] = tuple()


def rgetattr(obj: object, attr: str, default: object = None) -> object:
    """Get nested attributes, recursively

    Usage:
        >> repr(my_object)
            Object(hello=Object(world=2))
        >> rgetattr(my_object, "hello.world")
            2
        >> rgetattr(my_object, "hello.world.space")
            None
        >> rgetattr(my_object, "hello.world.space", 20)
            20
    """
    result = _rgetattr(obj, attr)
    return default if result is _SENTINEL else result


def _rgetattr(obj: object, attr: str) -> object:
    """Get nested attributes, recursively"""

    def _getattr(obj, attr):
        return getattr(obj, attr, _SENTINEL)

    return functools.reduce(_getattr, [obj] + attr.split("."))  # type: ignore


def clean_completion_name(name: str, char: str) -> str:
    """Clean the completion name, stripping bad surroundings

    1. Remove all surrounding " and '. For
    """
    if char == "'":
        return name.lstrip("'")
    if char == '"':
        return name.lstrip('"')
    return name


def char_before_cursor(
    document: Document, position: Position, default=""
) -> str:
    """Get the character directly before the cursor"""
    try:
        return document.lines[position.line][position.character - 1]
    except IndexError:
        return default


def uuid() -> str:
    """Function to create a string uuid"""
    return str(uuid4())
